#include <inarm.h>
#include <stdio.h>
#include <stdlib.h>
#include "board.h"

// Intel Basic Command Set Commands
#define	READ_ID			0x90
#define PROGRAM_COMMAND		0x40		//or 0x10 both are valid
#define PROGRAM_VERIFY		0xD0
#define READ_STATUS		0x70
#define CLEAR_STATUS		0x50
#define READ_ARRAY		0xFF
#define BLOCK_ERASE		0x20
#define	PROG_ERASE_SUSPEND	0xB0
#define	PPROG_ERASE_RESUME	0xD0
#define BLOCK_LOCK_BITS		0x60
#define	PROT_PRG		0xC0
#define BLOCK_LOCK_CONFIRM	0x01
#define CFI_QUERY_COMMAND	0x98
#define CFI_DATA_OFFS		0x20
#define RESET_DEV		0xF0

// Word access
#define FIRST_KEY_ADD   	(0x555<<1)
#define SECOND_KEY_ADD   	(0x2AA<<1)
#define THIRD_KEY_ADD   	(0x555<<1)

#define KEY_AA           	0xAA
#define KEY_55           	0x55

#define BIT_PASS_W              0x00000080

// Intel Status masks
#define STATUS_LOCKED		0x02
#define LOCK_ERROR		0x10
#define ERASE_ERROR		0x20
#define STATUS_READY		0x80


#define FLASH_BEGIN             0x40000000
#define FLASH_BOOT              0x8000
// #define FLASH_BOOT              0
#define FLASH_SIZE              (0x80000-FLASH_BOOT)   // Total area of Flash region in words 8 bit
#define FLASH_BASE              (FLASH_BEGIN+FLASH_BOOT)   //First 0x8000 bytes reserved for boot loader etc.

#define FLASH_ERROR             0x01

#define FLASH_OK      0
#define FLASH_ERR     1

// variables
unsigned short status;
FLASH_DEV_MANUFACTURER_ID_DEF FlashDevManID = UNKNOW_MAN_ID;
unsigned int id;
unsigned long* pAdd;
unsigned short Data;
unsigned int error_status;
unsigned long hold;

// write command at specify address
int command(unsigned char command, unsigned short *address) {

  *address = (unsigned short)command;
  return PASS;
}


// This checks the status register, and loops untill the Flash device is 'ready'
int ReadyWait(unsigned short* address,unsigned short data, FLASH_DEV_MANUFACTURER_ID_DEF DevID) {

  if (DevID == INTEL_MAN_ID)
  {
    while((*address & STATUS_READY)!=STATUS_READY);
    return PASS;
  }
  else if (DevID == MXID_MAN)
  {
    while ((data ^ *address) & BIT_PASS_W);
    if ((data ^ *address) & BIT_PASS_W)
      return(FAIL);
    return(PASS);
  }
  return(FALSE);
}

// read device ID
int FlashID(unsigned short* address)
{
  unsigned int hold;

  // Put device into 'READ_ID'.
  command(KEY_AA,(unsigned short*)(FIRST_KEY_ADD+FLASH_BEGIN));
  command(KEY_55,(unsigned short*)(SECOND_KEY_ADD+FLASH_BEGIN));
  command(READ_ID,(unsigned short*)(THIRD_KEY_ADD+FLASH_BEGIN));

  hold = *address;                // Read Data

  command (RESET_DEV, address);   // Reset 26LVXXXX
  command (READ_ARRAY, address);  // C3 Put device back into read mode.

  // Return Item
  return(hold);
}


// Unlock block flash
int Flash_Unlock_Block(unsigned short *address, FLASH_DEV_MANUFACTURER_ID_DEF DevID) {

  command (READ_ARRAY, address);		// Put device back into read mode.
  command (BLOCK_LOCK_BITS, address);   	// Clear lock bits write 0x60 within the device address space
  command (PROGRAM_VERIFY, address);    	// Confirm action by writing 0xD0 within the device address space
  command (READ_ARRAY, address);		// Put device back into read mode.
  return PASS;
}


// This routine writes a single 32 bit word to Flash.  On entry, it assumes the
// Flash is in READ-ARRAY mode (which it is following reset, or exit from any other
// non-static function in this source file.
// For 16-bit access flash, Little Endian is presumed.
// This function checks whether the block is locked before attempting the write.
int Flash_Write_Word(unsigned long *address, unsigned long data, FLASH_DEV_MANUFACTURER_ID_DEF DevID) {

  int i;
  unsigned short shdata;
  shdata=(short)data;

  if(DevID == INTEL_MAN_ID) {
    for (i=0;i<2;i++) {
      command(PROGRAM_COMMAND,(unsigned short*)address);	      // Writing 0x40 puts devices into write mode.
      *((volatile unsigned short*)address+i) = shdata;                // Write data
      command (READ_STATUS,(unsigned short*)address);	              // Put device into 'READ STATUS MODE'.
      ReadyWait((unsigned short*)address,0,INTEL_MAN_ID);	      // Wait for write to complete.
      while(*(volatile unsigned short*)address & STATUS_LOCKED)       // Block was locked (if should suffice, but while good for debug).
      {
        Flash_Unlock_Block((unsigned short*)address,INTEL_MAN_ID);    // re-issue program command 0x10 (word)
        command(CLEAR_STATUS, (unsigned short*)address);              // clear status error
        command(PROGRAM_COMMAND, (unsigned short*)address+i);         // address simply needs to be in the correct device
        *((volatile unsigned short*)address+i) = shdata;	      // Write required word to required address
        command (READ_STATUS, (unsigned short*)address);	      // Put device into 'READ STATUS MODE'.
        ReadyWait((unsigned short*)address,0,INTEL_MAN_ID);		
      }
      shdata = data >> 16;
    }

    command(READ_ARRAY, (unsigned short*)address);	// Reset for Read operation
    if (*address != data)                               // Read word back to verify
      return FAIL;
  }
  else if (DevID == MXID_MAN)
  {
    for (i=0;i<2;i++)
    {
      // Flash Write Command
      command(KEY_AA,(unsigned short*)(FIRST_KEY_ADD+FLASH_BEGIN));
      command(KEY_55,(unsigned short*)(SECOND_KEY_ADD+FLASH_BEGIN));
      command(0xA0,(unsigned short*)(THIRD_KEY_ADD+FLASH_BEGIN));
      *((volatile unsigned short*)address+i) = shdata;	                    // Write required word to required address
      if(ReadyWait((unsigned short*)address+i,shdata,MXID_MAN) != PASS){    // Check Write End
        return FAIL;
      }
      shdata = data >> 16;
    }
  }
  else
    return FALSE;
  return PASS;
}

// This function writes an arbitary area of flash.
// It first verfies that the data will fit within the flash address space,
// Then divides the operation into calls to Flash_Write_Word, which writes
// data areas contained within a single block.
int Flash_Write_Area(unsigned long *address, long *data, long size, FLASH_DEV_MANUFACTURER_ID_DEF DevID) {

  if ((address >= (unsigned long*)FLASH_BASE) && ((address+size/4) <= (unsigned long*)(FLASH_BASE + FLASH_SIZE)))
  {
    for (unsigned i = 0; i!=size; ++i)
    {
      if(Flash_Write_Word(address,*(data+i),DevID)!=PASS) return FAIL;
    }
  }
  else
  {
    return FAIL;
  }
  return PASS;
}


// This function erases a single block, specified by it's base address
int Flash_Erase_Block(unsigned long *address, FLASH_DEV_MANUFACTURER_ID_DEF DevID) {

  int result;
  unsigned int s_reg;

  if (DevID == INTEL_MAN_ID)
  {
    Flash_Unlock_Block((unsigned short*)address,INTEL_MAN_ID);
    do
    {
      result = PASS;
      command(BLOCK_ERASE, (unsigned short*)address);	        // Issue single block erase command
      command(PROGRAM_VERIFY, (unsigned short*)address);	// Confirm Action 0xd0 with block address
      command(READ_STATUS, (unsigned short*)address);
      ReadyWait((unsigned short*)address,0xFFFF,INTEL_MAN_ID);	// Wait for erase attempt to complete
      s_reg = *(volatile unsigned int*)address;

      // This results in reading s_reg twice (first time in readywait), so Intel errata does not cause problem.
      if (s_reg & (ERASE_ERROR | (ERASE_ERROR<<16)))
      {
        command(CLEAR_STATUS, (unsigned short*)address);
        result = FAIL;
      }
      // Reset for Read operation
      command(READ_ARRAY, (unsigned short*)address);
    } while (result == FAIL);
  }
  else
  {
    command(KEY_AA,(unsigned short*)(FIRST_KEY_ADD+FLASH_BEGIN));
    command(KEY_55,(unsigned short*)(SECOND_KEY_ADD+FLASH_BEGIN));
    command(0x80,(unsigned short*)(THIRD_KEY_ADD+FLASH_BEGIN));
    command(KEY_AA,(unsigned short*)(FIRST_KEY_ADD+FLASH_BEGIN));
    command(KEY_55,(unsigned short*)(SECOND_KEY_ADD+FLASH_BEGIN));
    command(0x30,(unsigned short*)address);
    result = ReadyWait((unsigned short*)address,0xFFFF,MXID_MAN);
  }
  return (result);
}

// This function erases all blocks overlapping specified region (via calls to Flash_Erase_Block)
int Flash_Erase_Blocks(unsigned long *address, long size, FLASH_DEV_MANUFACTURER_ID_DEF DevID) {

  unsigned long* limit = address+size/4-1;
  for(;address<=limit;address+=FLASH_BLOCK/4)
   Flash_Erase_Block(address,DevID);
  return PASS;
}



// test flash
char TestFlash(void) {

  unsigned long i = 0;

  // Flash memory test
  //printf("\n\rFlash Tests\n\r");
  //printf("Read manufacturer\n\r");


  UARTString("\n\r\n\r************ FLASH Test begin ************");

  UARTString("\n\r1. Read manufacturer - ");

  id = FlashID((unsigned short*)(FLASH_BASE)+MANUFACTURER_ID);
  if (id == 0x0089)
  {
    // printf("Intel flash device is find\n");
    UARTString("Intel flash device is find");

    // Set flash man. type
    FlashDevManID = INTEL_MAN_ID;
  }
  else if (id == 0x00C2)
  {
    // printf("MXIC flash device is find\n");
    UARTString("MXIC flash device is find");

    // Set flash man. type
    FlashDevManID = MXID_MAN;
  }
  else
  {
    // printf("%s %#x %c","Unknow flash device is find",id,'\n');
    UARTString("Unknow flash device is find");
  }

  // printf("Read type of chip\n");
  UARTString("\n\r2. Read type of chip - ");

  id = FlashID((unsigned short*)(FLASH_BASE)+DEVICE_ID);
  if (id == 0x88C5)
  {
    // printf("TE28F320C3Bxxx\n");
    UARTString("TE28F320C3Bxxx");
  }
  else if (id == 0x225B)
  {
    // printf("TE28F320C3Bxxx\n");
    UARTString("TE28F320C3Bxxx");
  }
  else if (id == 0x88C4)
  {
    // printf("TE28F320C3Bxxx\n");
    UARTString("TE28F320C3Bxxx");
  }
  else
  {
    // printf("%s %#x %c","Unknow flash device is find",id,'\n');
    UARTString("Unknow flash device is find");

    FlashDevManID = UNKNOW_MAN_ID;
  }

  if (FlashDevManID != UNKNOW_MAN_ID)
  {
    // printf("Flash Erase\n");
    UARTString("\n\r3. Flash Erase - ");

    for(i=8, pAdd=(unsigned long *)FLASH_BASE; pAdd < (unsigned long *)FLASH_BASE+FLASH_SIZE/4; pAdd += FLASH_BLOCK/4,++i)
    {
      // LPC210xLedSet(++j&1);
      if (Flash_Erase_Block(pAdd,FlashDevManID) != PASS)
        break;
    }
    if (pAdd >= (unsigned long *)FLASH_BASE+FLASH_SIZE/4)
    {
      // printf("%s","Flash Erase Pass\n");
      UARTString("Flash Erase Pass");
    }
    else
    {
      // printf("%s %d %c","Erase Block - ",i,'\n');
      // printf("%s","Flash Erase fault\n");
      UARTString("Flash Erase fault");


      error_status |= FLASH_ERROR;

      return FLASH_ERR;
    }

    if ((error_status &  FLASH_ERROR) == 0)
    {
      // printf("%s","Flash Write\n");
      UARTString("\n\r4. Flash Write - ");

      for(i=0, pAdd=(unsigned long *)FLASH_BASE; i < FLASH_SIZE/4; ++pAdd,++i)
      {
        // if ((i & 0x1FFF) == 0)
          //LPC210xLedSet(++j&1);
        if (Flash_Write_Word(pAdd,i,FlashDevManID) != PASS)
          break;
      }

      if (pAdd >= (unsigned long *)FLASH_BASE+FLASH_SIZE/4)
      {
        // printf("%s","Flash Write Pass\n");
        UARTString("Flash Write Pass");
      }
      else
      {
        // printf("%s %#x %c","Flash Write Fault at address - ",pAdd,'\n');
        UARTString("Flash Write Fault");

        error_status |= FLASH_ERROR;

        return FLASH_ERR;
      }

      if ((error_status &  FLASH_ERROR) == 0)
      {
        // printf("%s","Flash Verify Data\n");
        UARTString("\n\r5. Flash Verify Data - ");

        for(i=0, pAdd=(unsigned long *)FLASH_BASE; i < FLASH_SIZE/4; ++pAdd,++i)
        {
          hold = *pAdd;
          if (hold != i)
            break;
          //if ((i & 0x1FFF) == 0)
          //  LPC210xLedSet(++j&1);
        }

        if (pAdd >= (unsigned long *)FLASH_BASE+FLASH_SIZE/4)
        {
          // printf("%s","Flash Verify Pass\n");
          UARTString("Flash Verify Pass");
        }
        else
        {
          // printf("%s %#x %c","Flash Verify Fault at address - ",pAdd,'\n');
          UARTString("Flash Verify Fault");

          error_status |= FLASH_ERROR;

          return FLASH_ERR;
        }
      }
    }
  }
  else
  {
    // printf("Unknow type flash device is find. Test fault");
    UARTString("Unknow type flash device is find. Test fault");

    error_status |= FLASH_ERROR;
    return FLASH_ERR;
  }

  return FLASH_OK;
}

