// mmc.c : MultiMediaCard functions: init, read, write ...
//
// Rolf Freitag 5/2003
//



// MMC Lib
#ifndef _MMCLIB_C
#define _MMCLIB_C
//---------------------------------------------------------------------

#include  "board.h"
#include  "math.h"
#include  "string.h"

CSP_SPI_T *   const SPI0  = (CSP_SPI_T *)SPI0_BASE_ADDRESS;
CSP_PIO_T *   const UPIO  = (CSP_PIO_T *)UPIO_BASE_ADDRESS;
CSP_PMC_T *   const PMC   = (CSP_PMC_T *)PMC_BASE_ADDRESS;
CSP_PDC_T *   const PDC   = (CSP_PDC_T *)PDC_BASE_ADDRESS;

extern CSP_PIO_T * const UPIO;

char mmcGetResponse(void);
char mmcGetXXResponse(const char resp);
char mmcCheckBusy(void);

void initSSP (void);

char mmc_buffer_test_1[512];
char mmc_buffer_test_2[512];

char mmc_buffer[512] = { 0 };	// Buffer for mmc i/o for data and registers
extern char card_state;		// 0 for no card found, 1 for card found (init successfull)


//---------------------------------------------------------------------
void Delays (unsigned long a) { while (--a!=0); }


void initSPI_port (void) {

  // Enable clock for PIO
  UPIO->ECR = 0x1;

  //Card present -> CP - UPIO_1
  UPIO->ODR |= BIT1; //Configure in Input
  //UPIO->PER = BIT1; //Enable UPIO_1

  //Write protect -> WP - UPIO_0
  UPIO->ODR |= BIT0; //Configure in Input
  //UPIO->PER = BIT0; //Enable UPIO_0


  UPIO->ODR = 0xffffffff;
}

// setup usart1 in spi mode
void initSPI (void)
{

  // enables the clock
  PMC->ECR = PMC_ARM|PMC_PDC;
//  PMC->ECR = PMC_PDC;


  //SPI Enable, Sowtware reset
  SPI0->CR = SWRST;

  // set functionalite to pins:
  // SPI0->PER = SPCK | MISO | MOSI | NPCS1;
  SPI0->PDR = SPCK | MISO | MOSI | NPCS0 | NPCS1 | NPCS2 | NPCS3;
  // SPI0->OER = SPCK | MOSI | NPCS1;

  // enable the clock of SPI
  SPI0->ECR = SPI|PIO;

  // Master mode, fixed select, disable decoder, FDIV=0 (MCK), PCS=1110
  SPI0->MR = MSTR | PCS1;

  //8bit, CPOL=0, ClockPhase=1, SCLK = 200kHz
  SPI0->CS[1].CSR  = NCPHA + SCBR(0x55);

    //SPI Enable
  SPI0->CR = SPIEN;

}


// Initialisieren
char initMMC (void)
{

  //raise SS and MOSI for 80 clock cycles
  //SendByte(0xff) 10 times with SS high
  //RAISE SS
  int i;
  char response=0x01;


 // debug_printf("Start iniMMC......");
  initSPI();

  //initialization sequence on PowerUp
  ///CS_HIGH();
  for(i=0;i<=9;i++)
    spiSendByte(0xff);
  ///CS_LOW();
  //Send Command 0 to put MMC in SPI mode
  mmcSendCmd(0x00,0,0x95);
  //Now wait for READY RESPONSE
  if(mmcGetResponse()!=0x01);
  //  debug_printf("no responce");

  while(response==0x01)
  {
    //  debug_printf("Sending Command 1");
   //CS_HIGH();
   spiSendByte(0xff);
   //CS_LOW();
   mmcSendCmd(0x01,0x00,0xff);
   response=mmcGetResponse();
  }
  //CS_HIGH();
  spiSendByte(0xff);
 // debug_printf("MMC INITIALIZED AND SET TO SPI MODE PROPERLY.");
  return MMC_SUCCESS;
}



// Ti added mmc Get Responce
char mmcGetResponse(void)
{
  //Response comes 1-8bytes after command
  //the first bit will be a 0
  //followed by an error code
  //data will be 0xff until response
  int i=0;

  char response;

  while(i<=64)
  {
   response=spiSendByte(0xff);
   if(response==0x00)break;
   if(response==0x01)break;
   i++;
  }
  return response;
}

char mmcGetXXResponse(const char resp)
{
  //Response comes 1-8bytes after command
  //the first bit will be a 0
  //followed by an error code
  //data will be 0xff until response
  int i=0;

  char response;

  while(i<=500)
  {
   response=spiSendByte(0xff);
   if(response==resp)break;
   i++;
  }
  return response;
}
char mmcCheckBusy(void)
{
  //Response comes 1-8bytes after command
  //the first bit will be a 0
  //followed by an error code
  //data will be 0xff until response
  int i=0;

  char response;
  char rvalue;
  while(i<=64)
  {
   response=spiSendByte(0xff);
   response &= 0x1f;
   switch(response)
    {
     case 0x05: rvalue=MMC_SUCCESS;break;
     case 0x0b: return(MMC_CRC_ERROR);
     case 0x0d: return(MMC_WRITE_ERROR);
     default:
	      rvalue = MMC_OTHER_ERROR;
	      break;
    }
   if(rvalue==MMC_SUCCESS)break;
    i++;
  }
  i=0;
  do
  {
   response=spiSendByte(0xff);
   i++;
  }while(response==0);
  return response;
}
// The card will respond with a standard response token followed by a data
// block suffixed with a 16 bit CRC.

// Ti Modification: long int -> long ; int -> long
char mmcReadBlock(const unsigned long address, const unsigned long count)
{
  unsigned long i = 0;
  char rvalue = MMC_RESPONSE_ERROR;

  // Set the block length to read
  if (mmcSetBlockLength (count) == MMC_SUCCESS)	// block length could be set
    {
      //SS = LOW (on)
      //CS_LOW ();
      // send read command MMC_READ_SINGLE_BLOCK=CMD17
      mmcSendCmd (17,address, 0xFF);
      // Send 8 Clock pulses of delay, check if the MMC acknowledged the read block command
      // it will do this by sending an affirmative response
      // in the R1 format (0x00 is no errors)
      if (mmcGetResponse() == 0x00)
	{
	  // now look for the data token to signify the start of
	  // the data
	  if (mmcGetXXResponse(MMC_START_DATA_BLOCK_TOKEN) == MMC_START_DATA_BLOCK_TOKEN)
	    {
	      // clock the actual data transfer and receive the bytes; spi_read automatically finds the Data Block
	      for (i = 0; i < 512; i++)
		mmc_buffer[i] = spiSendByte(0xff);	// is executed with card inserted

	      // get CRC bytes (not really needed by us, but required by MMC)
	      spiSendByte(0xff);
	      spiSendByte(0xff);
	      rvalue = MMC_SUCCESS;
	    }
	  else
	    {
	      // the data token was never received
	      rvalue = MMC_DATA_TOKEN_ERROR;	// 3
	    }
	}
      else
	{
	  // the MMC never acknowledge the read command
	  rvalue = MMC_RESPONSE_ERROR;	// 2
	}
    }
  else
    {
      rvalue = MMC_BLOCK_SET_ERROR;	// 1
    }
  //CS_HIGH ();
  spiSendByte(0xff);
  return rvalue;
}				// mmc_read_block



//---------------------------------------------------------------------
// Ti Modification: long int -> long
char mmcWriteBlock (const unsigned long address)
{
  unsigned long i = 0;
  char rvalue = MMC_RESPONSE_ERROR;	// MMC_SUCCESS;
  //char c = 0x00;

  // Set the block length to read
  if (mmcSetBlockLength (512) == MMC_SUCCESS)	// block length could be set
    {
      // SS = LOW (on)
      //CS_LOW ();
      // send write command
      mmcSendCmd (24,address, 0xFF);

      // check if the MMC acknowledged the write block command
      // it will do this by sending an affirmative response
      // in the R1 format (0x00 is no errors)
      if (mmcGetXXResponse(MMC_R1_RESPONSE) == MMC_R1_RESPONSE)
	{
	  spiSendByte(0xff);
	  // send the data token to signify the start of the data
	  spiSendByte(0xfe);
	  // clock the actual data transfer and transmitt the bytes
	  for (i = 0; i < 512; i++)
	    spiSendByte(mmc_buffer[i]);	// mmc_buffer[i];       Test: i & 0xff
	  // put CRC bytes (not really needed by us, but required by MMC)
	  spiSendByte(0xff);
	  spiSendByte(0xff);
	  // read the data response xxx0<status>1 : status 010: Data accected, status 101: Data
	  //   rejected due to a crc error, status 110: Data rejected due to a Write error.
          mmcCheckBusy();
	}
      else
	{
	  // the MMC never acknowledge the write command
	  rvalue = MMC_RESPONSE_ERROR;	// 2
	}
    }
  else
    {
      rvalue = MMC_BLOCK_SET_ERROR;	// 1
    }
  //give the MMC the required clocks to finish up what ever it needs to do
  //for (i = 0; i < 9; ++i)
  //spiSendByte(0xff);

  //CS_HIGH ();
  // Send 8 Clock pulses of delay.
  spiSendByte(0xff);
  return rvalue;
}// mmc_write_block


//---------------------------------------------------------------------
void mmcSendCmd (const char cmd, unsigned long data, const char crc)
{
  char frame[6];
  char temp;
  int i;

  frame[0]=(cmd|0x40);
  for(i=3;i>=0;i--){
    temp=(char)(data>>(8*i));
    frame[4-i]=(temp);
  }
  frame[5]=(crc);
  for(i=0;i<6;i++)
    spiSendByte(frame[i]);
}


//--------------- set blocklength 2^n ------------------------------------------------------
// Ti Modification: long int-> long
char mmcSetBlockLength (const unsigned long blocklength)
{
  //char rValue = MMC_TIMEOUT_ERROR;
  //char i = 0;

  // SS = LOW (on)
  //CS_LOW ();

  // Set the block length to read
  //MMC_SET_BLOCKLEN =CMD16
  mmcSendCmd(16, blocklength, 0xFF);

  // get response from MMC - make sure that its 0x00 (R1 ok response format)
  if(mmcGetResponse()!=0x00);

  //CS_HIGH ();

  // Send 8 Clock pulses of delay.
  spiSendByte(0xff);

  return MMC_SUCCESS;
}				


unsigned char spiSendByte(const unsigned char data)
{
    unsigned int spib;

///*
    while((SPI0->SR & TDRE) == 0);  // Wait for the transfer to complete
    // SPI0->TDR = (data & 0xFFFF);              // Send the data
    SPI0->TDR = data;

    while((SPI0->SR & RDRF) == 0);  // Wait until the character can be sent
    // spib = ((SPI0->RDR) & 0xFFFF);      // Get the data received
    spib = (SPI0->RDR);
    return spib;
//*/

//   /* PDC Definition */
//   /* Send to Peripheral Opcode to read to a specified address */
//   CSP_PDCInitTx((U32_T)SPI0, (U32_T)&(SPI0->TDR), SIZE_BYTE);
//   /* Receive data from SPI */
//   CSP_PDCInitRx((U32_T)SPI0, (U32_T)&(SPI0->RDR), SIZE_BYTE);
//
//   /* Enable SPI */
//   CSP_SPI_SET_CR(SPI0, SPIEN);
//
//   /* Start Dialog */
//   CSP_PDCStartTx((U32_T)SPI0, (U32_T)&data, 1);
//   CSP_PDCStartRx((U32_T)SPI0, (U32_T)&spib, 1);
//
//   return spib;
}



// Reading the contents of the CSD and CID registers in SPI mode is a simple
// read-block transaction.
char mmcReadRegister (const char cmd_register, const unsigned char length)
{
  char uc = 0;
  char rvalue = MMC_TIMEOUT_ERROR;
//  char i = 0;

  if (mmcSetBlockLength (length) == MMC_SUCCESS)
    {
      // CS_LOW ();
      // CRC not used: 0xff as last byte
      mmcSendCmd(cmd_register, 0x000000, 0xff);

      // wait for response
      // in the R1 format (0x00 is no errors)
      if (mmcGetResponse() == 0x00)
	{
	 if (mmcGetXXResponse(0xfe)== 0xfe)
	    for (uc = 0; uc < length; uc++)
	      mmc_buffer[uc] = spiSendByte(0xff);
	  // get CRC bytes (not really needed by us, but required by MMC)
	  spiSendByte(0xff);
	  spiSendByte(0xff);
	}
      else
	rvalue = MMC_RESPONSE_ERROR;
      //CS=HIGH (off)
      //CS_HIGH ();

      // Send 8 Clock pulses of delay.
      spiSendByte(0xff);
    }
  //CS_HIGH ();
  return rvalue;
}// mmc_read_register


unsigned char TestMMC(void) {

  char flag_cp=0;
  char flag_wp=0;

    // variable init
  memset(&mmc_buffer_test_1,'0',512);
  memset(&mmc_buffer_test_2,'1',512);

  UARTString("\n\r\n\r************ MMC Test begin ************");

  // check for CP and WP
  initSPI_port();

  // CP
  while((UPIO->PDSR&BIT1)!=BIT1) {

    if(flag_cp==0) {
       UARTString("\n\rCard is not present");
       flag_cp=1;
    }
  }

  // WP
  while((UPIO->PDSR&BIT0)!=BIT0) {

    if(flag_wp==0) {
       UARTString("\n\rCard is write protected");
       flag_wp=1;
    }
  }

  UARTString("\n\rCard is present, not protected and ready to test");

  /**** MMC CARD ****/
  if (initMMC() == MMC_SUCCESS)	// card found
  {
    memset(&mmc_buffer,0,512);
    mmcReadRegister (10, 16);
    mmc_buffer[7]=0;

    // Fill first Block (0) with 'A'
    memset(&mmc_buffer,'0',512);    //set breakpoint and trace mmc_buffer contents
    mmcWriteBlock(0);

    // Fill second Block (1)-AbsAddr 512 with 'B'
    memset(&mmc_buffer,'1',512);
    mmcWriteBlock(512);

    // Read first Block back to buffer
    memset(&mmc_buffer,0x00,512);
    mmcReadBlock(0,512);
    if(strncmp(&mmc_buffer[0], &mmc_buffer_test_1[0], 512)) return 1;

    // Read first Block back to buffer
    memset(&mmc_buffer,0x00,512);
    mmcReadBlock(512,512);
    if(strncmp(&mmc_buffer[0], &mmc_buffer_test_2[0], 512)) return 1;

    memset(&mmc_buffer,0x00,512);

    return 0;

  }

  return 1;
}

//---------------------------------------------------------------------
#endif /* _MMCLIB_C */
